mod game;

use axum::{
    Router,
    body::Bytes,
    extract::{
        ConnectInfo, State,
        ws::{Message, WebSocket, WebSocketUpgrade},
    },
    http::{Request, header},
    response::IntoResponse,
    routing::{any, get},
};
use core::{net::SocketAddr, str::FromStr};
use futures::{FutureExt, TryFutureExt};
use game::GameStateSender;
use scratchpad::{ServerMessage, game::TileType};
use std::{env, io::ErrorKind, sync::Arc};
use tokio::{
    io::{AsyncReadExt, AsyncWriteExt},
    sync::{
        Mutex,
        broadcast::{Receiver, error::RecvError},
    },
};

static GAME_STATE: Mutex<GameStateSender> = Mutex::const_new(GameStateSender::new());

struct AppState {
    recv: Receiver<Arc<[TileType]>>,
}
impl Clone for AppState {
    fn clone(&self) -> Self {
        Self {
            recv: self.recv.resubscribe(),
        }
    }
}

const DEFAULT_PORT: u16 = 8080;
const DEFAULT_HOST: &str = "127.0.0.1";
const GAME_STATE_SAVE_FILE_NAME: &str = "/tmp/scratchpad.state";

async fn game_saver(mut recv: Receiver<Arc<[TileType]>>) {
    loop {
        let board = match recv.recv().await {
            Ok(board) => board,
            Err(RecvError::Closed) => {
                panic!("save channel closed")
            }
            Err(err) => {
                log::error!("game_saver: {err}");
                continue;
            }
        };

        let mut save_file = match tokio::fs::File::create(GAME_STATE_SAVE_FILE_NAME).await {
            Ok(file) => file,
            Err(err) => {
                log::error!("opening save file: {err}");
                continue;
            }
        };
        let board_bytes = board.iter().map(|t| *t as u8).collect::<Box<_>>();
        if let Err(err) = save_file.write_all(&board_bytes).await {
            core::mem::drop(save_file);
            log::error!("writing save file: {err}");
            continue;
        }
        if let Err(err) = save_file.flush().await {
            core::mem::drop(save_file);
            log::error!("flushing save file: {err}");
            continue;
        }
        core::mem::drop(save_file);
    }
}

async fn restore_game_save_or_give_up() {
    let mut game_save = Vec::new();
    let save_ref = game_save.as_mut();

    let save_slice = match tokio::fs::File::open(GAME_STATE_SAVE_FILE_NAME)
        .and_then(|mut f| async move { f.read_to_end(save_ref).await })
        .await
    {
        Ok(read) => &game_save[..read],
        Err(err) => {
            log::error!("loading game save: {err}; using blank game state");
            return;
        }
    };

    if let Err(err) = GAME_STATE.lock().await.replace_game_state(save_slice) {
        log::error!("restoring game save: {err}; using blank game state")
    }
}

#[tokio::main]
async fn main() {
    pretty_env_logger::init();

    let (sender, recv) = tokio::sync::broadcast::channel::<Arc<[TileType]>>(512);
    restore_game_save_or_give_up().await;
    GAME_STATE.lock().await.set_sender(sender);
    let recv_saver = recv.resubscribe();
    let saver_handle = tokio::spawn(async move { game_saver(recv_saver).await });

    let host = env::var("HOST").unwrap_or(DEFAULT_HOST.to_string());
    let port = env::var("PORT")
        .map_err(|err| anyhow::anyhow!("{err}"))
        .map(|port_str| {
            port_str
                .parse::<u16>()
                .unwrap_or_else(|err| panic!("parse PORT={port_str} failed: {err}"))
        })
        .unwrap_or(DEFAULT_PORT);
    let listen_addr =
        SocketAddr::from_str(format!("{host}:{port}").as_str()).expect("invalid host/port");

    let state = AppState { recv };

    let app = Router::new()
        .route("/scratchpad", any(ws_handler))
        .with_state(state)
        .fallback(get(handle_http_static));
    let listener = tokio::net::TcpListener::bind(listen_addr).await.unwrap();
    log::info!("listening on {}", listener.local_addr().unwrap());
    axum::serve(
        listener,
        app.into_make_service_with_connect_info::<SocketAddr>(),
    )
    .await
    .unwrap();
    saver_handle.await.unwrap();
}

async fn handle_http_static(req: Request<axum::body::Body>) -> impl IntoResponse {
    use mime_sniffer::MimeTypeSniffer;
    const INDEX_FILE: &[u8] = include_bytes!("../../blog/dist/index.html");
    let path = req.uri().path();
    log::info!("path: {path}");

    blog_macros::include_dist!(DIST_FILES, "blog/dist");

    let file = if let Some(file) = DIST_FILES.iter().find_map(|(file_path, file)| {
        if *file_path == path {
            Some(*file)
        } else {
            None
        }
    }) {
        file
    } else {
        return (
            [(header::CONTENT_TYPE, "text/html".to_string())],
            INDEX_FILE,
        );
    };

    let mime = if path.ends_with(".js") {
        "text/javascript".to_string()
    } else if path.ends_with(".css") {
        "text/css".to_string()
    } else if path.ends_with(".wasm") {
        "application/wasm".to_string()
    } else {
        file.sniff_mime_type()
            .unwrap_or("application/octet-stream")
            .to_string()
    };

    ([(header::CONTENT_TYPE, mime)], file)
}

async fn ws_handler(
    ws: WebSocketUpgrade,
    ConnectInfo(addr): ConnectInfo<SocketAddr>,
    State(state): State<AppState>,
) -> impl IntoResponse {
    log::info!("{addr} connected.");
    // finalize the upgrade process by returning upgrade callback.
    // we can customize the callback by sending additional info such as address.
    ws.on_upgrade(move |socket| handle_socket(socket, addr, state.clone().recv))
}

async fn on_recv(
    socket: &mut WebSocket,
    msg: Option<Result<Message, axum::Error>>,
) -> Result<(), anyhow::Error> {
    let msg = match msg {
        Some(Ok(msg)) => msg,
        Some(Err(err)) => return Err(err.into()),

        None => {
            log::info!("no message");
            return Ok(());
        }
    };
    let bytes = match msg {
        Message::Binary(bytes) => bytes,
        Message::Text(_) => return Err(anyhow::anyhow!("text")),
        Message::Ping(_) => return Err(anyhow::anyhow!("ping")),
        Message::Pong(_) => return Err(anyhow::anyhow!("pong")),
        Message::Close(close_frame) => {
            return Err(anyhow::anyhow!("sent close frame: {close_frame:?}"));
        }
    };
    if bytes.is_empty() {
        return Err(anyhow::anyhow!("got an empty bytes message"));
    }
    let bytes = bytes.iter().as_slice();
    let msg_type = scratchpad::ClientMessageType::try_from(bytes[0])?;
    log::info!("got message type: {msg_type:?}");
    let message = msg_type.try_parse_message(&bytes[1..])?;
    match message {
        scratchpad::ClientMessage::GetBoard => {
            let response: Vec<u8> =
                ServerMessage::StateUpdate(GAME_STATE.lock().await.get_board_values()).into();
            socket
                .send(Message::Binary(Bytes::from_owner(response)))
                .await?;
        }
        scratchpad::ClientMessage::SendMove(send_move) => {
            let new_board_state_message: Vec<u8> = ServerMessage::StateUpdate({
                let mut state = GAME_STATE.lock().await;
                state
                    .set_tile(send_move.row, send_move.col, send_move.value)
                    .await?;
                state.get_board_values()
            })
            .into();
            socket
                .send(Message::Binary(Bytes::from_owner(new_board_state_message)))
                .await?;
        }
    }

    Ok(())
}

async fn on_state_update(
    socket: &mut WebSocket,
    state: Result<Arc<[bool]>, RecvError>,
) -> Result<(), anyhow::Error> {
    let state = match state {
        Ok(state) => state,
        Err(err) => {
            log::error!("recv error: {err}");
            return Ok(());
        }
    };
    let mut out_bytes = Vec::new();
    out_bytes.push(scratchpad::ServerMessageType::StateUpdate as u8);
    out_bytes.extend(state.iter().map(|t| *t as u8));
    socket
        .send(Message::Binary(Bytes::from_owner(out_bytes)))
        .await?;
    Ok(())
}

async fn handle_socket(
    mut socket: WebSocket,
    who: SocketAddr,
    mut recv: Receiver<Arc<[TileType]>>,
) {
    loop {
        if let Err(err) = tokio::select! {
            msg = socket.recv() => {
                log::info!("socket.recv");
                on_recv(&mut socket, msg).await
            },
            r = recv.recv() => {
                log::info!("state update");
                on_state_update(&mut socket, r).await
            }
        } {
            log::error!("[{who}] {err}");
            return;
        }
    }
}
