use web_sys::{wasm_bindgen::JsCast, HtmlImageElement};
use yew::prelude::*;

pub const BADGE_INFO_LIST: &[BadgeInfo] = &[
    BadgeInfo::new("/img/badges/puff.webp", Some("/"), "jigglypuff.club"),
    BadgeInfo::new(
        "/img/badges/bsky.png",
        Some("https://bsky.app/profile/jigglypuff.club"),
        "Bluesky",
    ),
    BadgeInfo::with_special(
        "/img/badges/pronouns.png",
        None,
        "Pronouns: he/him",
        Special::OnClickSwitch("/img/badges/pronouns-animated.webp"),
    ),
    BadgeInfo::new(
        "/img/badges/ublock.png",
        Some("https://github.com/gorhill/uBlock"),
        "Use uBlock Origin! (or ad-blockers in general)",
    ),
    BadgeInfo::new(
        "/img/badges/consent-o-matic.png",
        Some("https://consentomatic.au.dk/"),
        "Use consent-o-matic! (or similar plugins)",
    ),
    BadgeInfo::new(
        "/img/badges/wiki.png",
        Some("https://donate.wikimedia.org/"),
        "Donate to Wikipedia",
    ),
    BadgeInfo::with_special(
        "/img/badges/riir-animated.webp",
        None,
        "Rewrite it in Rust!",
        Special::OnClickSwitch("/img/badges/riir.png"),
    ),
    BadgeInfo::with_special(
        "/img/badges/bunny.png",
        None,
        "Best viewed on Bunny Browser!",
        Special::OnClickSwitch("/img/badges/bunny-animated.webp"),
    ),
    BadgeInfo::with_special(
        "/img/badges/protect.png",
        Some("https://transkidsdeservebetter.org/support"),
        "Protect Trans Kids!",
        Special::OnHoverSwitch("/img/badges/protect-animated.webp"),
    ),
];
#[derive(Debug, Clone, Copy, PartialEq, Eq)]
pub enum Special {
    OnClickSwitch(&'static str),
    OnHoverSwitch(&'static str),
}

#[derive(Clone, Debug, Eq, PartialEq, Properties)]
pub struct BadgeInfo {
    pub src: &'static str,
    pub url: Option<&'static str>,
    pub name: &'static str,
    pub special: Option<Special>,
}

impl BadgeInfo {
    pub const fn new(src: &'static str, url: Option<&'static str>, name: &'static str) -> Self {
        Self {
            src,
            url,
            name,
            special: None,
        }
    }
    pub const fn with_special(
        src: &'static str,
        url: Option<&'static str>,
        name: &'static str,
        special: Special,
    ) -> Self {
        Self {
            src,
            url,
            name,
            special: Some(special),
        }
    }

    pub fn create_badge_html(&self) -> Html {
        html! {
            <Badge src={self.src} url={self.url} name={self.name} special={self.special}/>
        }
    }
}

fn switch_badge(event: MouseEvent) {
    log::warn!("switch_badge");
    if let Some(target) = event
        .target()
        .and_then(|t| t.dyn_into::<HtmlImageElement>().ok())
    {
        let new_src = match target.get_attribute("swap") {
            Some(new_src) => new_src,
            None => return,
        };
        let old_src = target.src();
        target.set_src(&new_src);
        target.set_attribute("swap", &old_src).unwrap();
    }
}

#[function_component]
pub fn Badge(info: &BadgeInfo) -> Html {
    let img = match info.special.as_ref() {
        Some(Special::OnClickSwitch(on_click_switch)) => {
            html! {
                <img class="badge" draggable="false" swap={*on_click_switch} onclick={switch_badge} src={info.src} alt={info.name} title={info.name}/>
            }
        }
        Some(Special::OnHoverSwitch(on_hover_switch)) => {
            html! {
                <img draggable="false" class="badge" swap={*on_hover_switch} onmouseover={switch_badge} src={info.src} alt={info.name} title={info.name}/>
            }
        }
        None => html! {
            <img class="badge" draggable="false" src={info.src} alt={info.name} title={info.name}/>
        },
    };
    if let Some(url) = info.url.as_ref() {
        html! {
            <a href={*url} target="_blank" rel="noopener noreferrer">
                {img}
            </a>
        }
    } else {
        img
    }
}
