use thiserror::Error;

pub const ROWS: usize = 32;
pub const COLS: usize = 32;
pub type TileType = bool;

#[derive(Debug, Clone, Copy, Error)]
pub enum GameError {
    #[error("coordinates out of range")]
    CoordinatesOutOfRange,
    #[error("board values message has invalid length")]
    BoardValuesMessageLengthInvalid,
    #[error("board values contains illegal value")]
    BoardValuesIllegalValue,
}

pub type BoardValues = [[TileType; COLS]; ROWS];

pub struct GameState {
    board: Board,
}

struct Row([TileType; COLS]);
struct Board {
    rows: [Row; ROWS],
}

impl GameState {
    pub const fn blank() -> Self {
        Self {
            board: Board {
                rows: unsafe {
                    core::mem::transmute::<[u8; core::mem::size_of::<[Row; COLS]>()], [Row; COLS]>(
                        [0u8; core::mem::size_of::<[Row; COLS]>()],
                    )
                },
            },
        }
    }

    pub fn get_board_values(&self) -> BoardValues {
        let mut values = [[TileType::default(); COLS]; ROWS];
        values
            .iter_mut()
            .map(|row| row.iter_mut())
            .zip(self.board.rows.iter().map(|row| row.0.iter()))
            .for_each(|(dst, src)| dst.zip(src).for_each(|(dst, src)| *dst = *src));
        values
    }

    pub fn replace_game_state(&mut self, bytes: &[u8]) -> Result<(), GameError> {
        let board_values = board_values_from_bytes(bytes)?;
        self.board
            .rows
            .iter_mut()
            .map(|row| row.0.iter_mut())
            .zip(board_values.iter().map(|row| row.iter()))
            .for_each(|(dst, src)| dst.zip(src).for_each(|(dst, src)| *dst = *src));
        Ok(())
    }

    pub fn set_tile(&mut self, row: u32, col: u32, value: TileType) -> Result<(), GameError> {
        if row >= ROWS as u32 || col >= COLS as u32 {
            return Err(GameError::CoordinatesOutOfRange);
        }
        self.board.rows[row as usize].0[col as usize] = value;
        Ok(())
    }
}

pub fn board_values_from_bytes(bytes: &[u8]) -> Result<BoardValues, GameError> {
    if bytes.len() != (ROWS * COLS) {
        return Err(GameError::BoardValuesMessageLengthInvalid);
    }
    if bytes.iter().any(|b| *b > 1) {
        return Err(GameError::BoardValuesIllegalValue);
    }
    let mut out = [[TileType::default(); COLS]; ROWS];
    out.iter_mut()
        .flat_map(|tiles| tiles.iter_mut())
        .zip(bytes.iter())
        .for_each(|(out, inp)| *out = *inp == 1);

    Ok(out)
}
